import os
from string import Template
from instana import internal_tracer, eum_api_key as global_eum_api_key
from instana.log import logger

# Usage:
#
# from instana.helpers import eum_snippet
# meta_kvs = { 'userId': user.id }
# eum_snippet(meta=meta_kvs)


def eum_snippet(trace_id=None, eum_api_key=None, meta={}):
    """
    Return an EUM snippet for use in views, templates and layouts that reports
    client side metrics to Instana that will automagically be linked to the
    current trace.

    @param trace_id [optional] the trace ID to insert into the EUM string
    @param eum_api_key [optional] the EUM API key from your Instana dashboard
    @param meta [optional] optional additional KVs you want reported with the
                EUM metrics

    @return string
    """
    try:
        eum_file = open(os.path.dirname(__file__) + '/eum.js')
        eum_src = Template(eum_file.read())

        # Prepare the standard required IDs
        ids = {}
        ids['meta_kvs'] = ''

        current_ctx = internal_tracer.current_context()

        if trace_id or current_ctx:
            ids['trace_id'] = trace_id or current_ctx.trace_id
        else:
            # No trace_id passed in and tracer doesn't show an active span so
            # return nothing, nada & zip.
            return ''

        if eum_api_key:
            ids['eum_api_key'] = eum_api_key
        else:
            ids['eum_api_key'] = global_eum_api_key

        # Process passed in EUM 'meta' key/values
        for key, value in meta.items():
            ids['meta_kvs'] += ("'ineum('meta', '%s', '%s');'" % (key, value))

        return eum_src.substitute(ids)
    except Exception as e:
        logger.debug(e)
        return ''

def eum_test_snippet(trace_id=None, eum_api_key=None, meta={}):
    """
    Return an EUM snippet for use in views, templates and layouts that reports
    client side metrics to Instana that will automagically be linked to the
    current trace.

    @param trace_id [optional] the trace ID to insert into the EUM string
    @param eum_api_key [optional] the EUM API key from your Instana dashboard
    @param meta [optional] optional additional KVs you want reported with the
                EUM metrics

    @return string
    """

    try:
        eum_file = open(os.path.dirname(__file__) + '/eum_test.js')
        eum_src = Template(eum_file.read())

        # Prepare the standard required IDs
        ids = {}
        ids['meta_kvs'] = ''

        current_ctx = internal_tracer.current_context()

        if trace_id or current_ctx:
            ids['trace_id'] = trace_id or current_ctx.trace_id
        else:
            # No trace_id passed in and tracer doesn't show an active span so
            # return nothing, nada & zip.
            return ''

        if eum_api_key:
            ids['eum_api_key'] = eum_api_key
        else:
            ids['eum_api_key'] = global_eum_api_key

        # Process passed in EUM 'meta' key/values
        for key, value in meta.items():
            ids['meta_kvs'] += ("'ineum('meta', '%s', '%s');'" % (key, value))

        return eum_src.substitute(ids)
    except Exception as e:
        logger.debug(e)
        return ''



import time
from basictracer import BasicTracer
import instana.recorder as r
import opentracing as ot
import instana
import instana.options as o

from instana.util import generate_id
from instana.span import InstanaSpan
from basictracer.context import SpanContext
from instana.http_propagator import HTTPPropagator
from instana.text_propagator import TextPropagator


class InstanaTracer(BasicTracer):
    sensor = None
    cur_ctx = None

    def __init__(self, options=o.Options()):
        self.sensor = instana.global_sensor
        super(InstanaTracer, self).__init__(
            r.InstanaRecorder(self.sensor), r.InstanaSampler())

        self._propagators[ot.Format.HTTP_HEADERS] = HTTPPropagator()
        self._propagators[ot.Format.TEXT_MAP] = TextPropagator()

    def start_span(
            self,
            operation_name=None,
            child_of=None,
            references=None,
            tags=None,
            start_time=None):
        "Taken from BasicTracer so we can override generate_id calls to ours"

        start_time = time.time() if start_time is None else start_time

        # See if we have a parent_ctx in `references`
        parent_ctx = None
        if child_of is not None:
            parent_ctx = (
                child_of if isinstance(child_of, ot.SpanContext)
                else child_of.context)
        elif references is not None and len(references) > 0:
            # TODO only the first reference is currently used
            parent_ctx = references[0].referenced_context

        # Assemble the child ctx
        instana_id = generate_id()
        ctx = SpanContext(span_id=instana_id)
        if parent_ctx is not None:
            if parent_ctx._baggage is not None:
                ctx._baggage = parent_ctx._baggage.copy()
            ctx.trace_id = parent_ctx.trace_id
            ctx.sampled = parent_ctx.sampled
        else:
            ctx.trace_id = instana_id
            ctx.sampled = self.sampler.sampled(ctx.trace_id)

        # Tie it all together
        span = InstanaSpan(
            self,
            operation_name=operation_name,
            context=ctx,
            parent_id=(None if parent_ctx is None else parent_ctx.span_id),
            tags=tags,
            start_time=start_time)

        self.cur_ctx = span.context
        return span

    def current_context(self):
        context = None
        if self.cur_ctx is not None:
            context = self.cur_ctx
        return context

    def inject(self, span_context, format, carrier):
        if format in self._propagators:
            self._propagators[format].inject(span_context, carrier)
        else:
            raise ot.UnsupportedFormatException()

    def extract(self, format, carrier):
        if format in self._propagators:
            return self._propagators[format].extract(carrier)
        else:
            raise ot.UnsupportedFormatException()

    def handle_fork(self):
        self.sensor.handle_fork()
        self.recorder = r.InstanaRecorder(self.sensor)


def init(options):
    """
    Deprecated.
        No longer in use.
        To be removed in next major release.
    """
    return instana.internal_tracer



import json
from instana import log
import instana.fsm as f
import instana.agent_const as a
import threading
from datetime import datetime

try:
    import urllib.request as urllib2
except ImportError:
    import urllib2


class From(object):
    pid = ""
    agentUuid = ""

    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class Head(urllib2.Request):

    def get_method(self):
        return "HEAD"


class Put(urllib2.Request):

    def get_method(self):
        return "PUT"


class Agent(object):
    sensor = None
    host = a.AGENT_DEFAULT_HOST
    port = a.AGENT_DEFAULT_PORT
    fsm = None
    from_ = From()
    last_seen = None

    def __init__(self, sensor):
        log.debug("initializing agent")

        self.sensor = sensor
        self.fsm = f.Fsm(self)

    def to_json(self, o):
        try:
            return json.dumps(o, default=lambda o: {k.lower(): v for k, v in o.__dict__.items()},
                              sort_keys=False, separators=(',', ':')).encode()
        except Exception as e:
            log.info("to_json: ", e, o)

    def is_timed_out(self):
        if self.last_seen and self.can_send:
            diff = datetime.now() - self.last_seen
            if diff.seconds > 60:
                return True
        return False

    def can_send(self):
        return self.fsm.fsm.current == "good2go"

    def head(self, url):
        return self.request(url, "HEAD", None)

    def request(self, url, method, o):
        return self.full_request_response(url, method, o, False, "")

    def request_response(self, url, method, o):
        return self.full_request_response(url, method, o, True, "")

    def request_header(self, url, method, header):
        return self.full_request_response(url, method, None, False, header)

    def full_request_response(self, url, method, o, body, header):
        b = None
        h = None
        try:
            if method == "HEAD":
                request = Head(url)
            elif method == "GET":
                request = urllib2.Request(url)
            elif method == "PUT":
                request = Put(url, self.to_json(o))
                request.add_header("Content-Type", "application/json")
            else:
                request = urllib2.Request(url, self.to_json(o))
                request.add_header("Content-Type", "application/json")

            response = urllib2.urlopen(request, timeout=2)

            if not response:
                self.reset()
            else:
                if response.getcode() < 200 or response.getcode() >= 300:
                    log.error("Request returned erroneous code", response.getcode())
                    if self.can_send():
                        self.reset()
                else:
                    self.last_seen = datetime.now()
                    if body:
                        b = response.read()

                    if header:
                        h = response.info().get(header)

                    if method == "HEAD":
                        b = True
        except Exception as e:
            # No need to show the initial 404s or timeouts.  The agent
            # should handle those correctly.
            if not (type(e) is urllib2.HTTPError and e.code == 404):
                log.debug("%s: full_request_response: %s" %
                          (threading.current_thread().name, str(e)))

        return (b, h)

    def make_url(self, prefix):
        return self.make_host_url(self.host, prefix)

    def make_host_url(self, host, prefix):
        port = self.sensor.options.agent_port
        if port == 0:
            port = a.AGENT_DEFAULT_PORT

        return self.make_full_url(host, port, prefix)

    def make_full_url(self, host, port, prefix):
        s = "http://%s:%s%s" % (host, str(port), prefix)
        if self.from_.pid != 0:
            s = "%s%s" % (s, self.from_.pid)

        return s

    def reset(self):
        self.last_seen = None
        self.from_ = From()
        self.fsm.reset()

    def set_host(self, host):
        self.host = host

    def set_port(self, port):
        self.port = port

    def set_from(self, json_string):
        if type(json_string) is bytes:
            raw_json = json_string.decode("UTF-8")
        else:
            raw_json = json_string

        self.from_ = From(**json.loads(raw_json))



import opentracing as ot
from instana import internal_tracer
import opentracing.ext.tags as tags


class iWSGIMiddleware(object):
    """ Instana WSGI middleware """

    def __init__(self, app):
        self.app = app
        self

    def __call__(self, environ, start_response):
        env = environ

        def new_start_response(status, headers, exc_info=None):
            """Modified start response with additional headers."""
            internal_tracer.inject(span.context, ot.Format.HTTP_HEADERS, headers)
            res = start_response(status, headers, exc_info)

            sc = status.split(' ')[0]
            if 500 <= int(sc) <= 511:
                span.set_tag("error", True)
                ec = span.tags.get('ec', 0)
                span.set_tag("ec", ec+1)

            span.set_tag(tags.HTTP_STATUS_CODE, sc)
            span.finish()
            return res

        if 'HTTP_X_INSTANA_T' in env and 'HTTP_X_INSTANA_S' in env:
            ctx = internal_tracer.extract(ot.Format.HTTP_HEADERS, env)
            span = internal_tracer.start_span("wsgi", child_of=ctx)
        else:
            span = internal_tracer.start_span("wsgi")

        span.set_tag(tags.HTTP_URL, env['PATH_INFO'])
        span.set_tag("http.params", env['QUERY_STRING'])
        span.set_tag(tags.HTTP_METHOD, env['REQUEST_METHOD'])
        span.set_tag("http.host", env['HTTP_HOST'])

        return self.app(environ, new_start_response)


def make_middleware(app=None, *args, **kw):
    """ Given an app, return that app wrapped in iWSGIMiddleware """
    app = iWSGIMiddleware(app, *args, **kw)
    return app



from __future__ import absolute_import

import opentracing as ot
from basictracer.context import SpanContext
from instana import util, log

prefix_tracer_state = 'X-INSTANA-'
prefix_baggage = 'X-INSTANA-BAGGAGE-'
field_name_trace_id = prefix_tracer_state + 'T'
field_name_span_id = prefix_tracer_state + 'S'


class TextPropagator():
    """
    A Propagator for TEXT_MAP.
    """

    def inject(self, span_context, carrier):
        try:
            carrier[field_name_trace_id] = '{0:x}'.format(span_context.trace_id)
            carrier[field_name_span_id] = '{0:x}'.format(span_context.span_id)
            if span_context.baggage is not None:
                for k in span_context.baggage:
                    carrier[prefix_baggage+k] = span_context.baggage[k]
        except Exception as e:
            log.debug("inject error: ", str(e))

    def extract(self, carrier):  # noqa
        try:
            if type(carrier) is dict or hasattr(carrier, "__dict__"):
                dc = carrier
            elif type(carrier) is list:
                dc = dict(carrier)
            else:
                raise ot.SpanContextCorruptedException()

            if field_name_trace_id in dc and field_name_span_id in dc:
                trace_id = util.header_to_id(dc[field_name_trace_id])
                span_id = util.header_to_id(dc[field_name_span_id])

            return SpanContext(span_id=span_id,
                               trace_id=trace_id,
                               baggage={},
                               sampled=True)

        except Exception as e:
            log.debug("extract error: ", str(e))
            return SpanContext()



from basictracer.span import BasicSpan
from basictracer.context import SpanContext


class InstanaSpan(BasicSpan):
    def finish(self, finish_time=None):
        if self.parent_id is None:
            self.tracer.cur_ctx = None
        else:
            # Set tracer context to the parent span
            pctx = SpanContext(span_id=self.parent_id,
                               trace_id=self.context.trace_id,
                               baggage={},
                               sampled=True)
            self.tracer.cur_ctx = pctx
        super(InstanaSpan, self).finish(finish_time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from flask import Flask, redirect
app = Flask(__name__)
app.debug = False
app.use_reloader = False


@app.route("/")
def hello():
    return "<center><h1>🐍 Hello Stan! 🦄</h1></center>"


@app.route("/301")
def threehundredone():
    return redirect('/', code=301)


@app.route("/302")
def threehundredtwo():
    return redirect('/', code=302)


@app.route("/400")
def fourhundred():
    return "Simulated Bad Request", 400


@app.route("/405")
def fourhundredfive():
    return "Simulated Method not allowed", 405


@app.route("/500")
def fivehundred():
    return "Simulated Internal Server Error", 500


@app.route("/504")
def fivehundredfour():
    return "Simulated Gateway Timeout", 504


@app.route("/exception")
def exception():
    raise Exception('fake error')


if __name__ == '__main__':
    app.run()



from __future__ import print_function
import opentracing as ot
from instana import internal_tracer
import opentracing.ext.tags as ext
import os


DJ19_INSTANA_MIDDLEWARE = 'instana.django19.InstanaMiddleware19'


class InstanaMiddleware19(object):
    """ Django 1.9 Middleware to provide request tracing for Instana """
    def __init__(self):
        self.span = None
        self

    def process_request(self, request):
        env = request.environ
        if 'HTTP_X_INSTANA_T' in env and 'HTTP_X_INSTANA_S' in env:
            ctx = internal_tracer.extract(ot.Format.HTTP_HEADERS, env)
            span = internal_tracer.start_span("django", child_of=ctx)
        else:
            span = internal_tracer.start_span("django")

        span.set_tag(ext.HTTP_URL, env['PATH_INFO'])
        span.set_tag("http.params", env['QUERY_STRING'])
        span.set_tag(ext.HTTP_METHOD, request.method)
        span.set_tag("http.host", env['HTTP_HOST'])
        self.span = span

    def process_response(self, request, response):
        if self.span:
            if 500 <= response.status_code <= 511:
                self.span.set_tag("error", True)
                ec = self.span.tags.get('ec', 0)
                self.span.set_tag("ec", ec+1)

            self.span.set_tag(ext.HTTP_STATUS_CODE, response.status_code)
            internal_tracer.inject(self.span.context, ot.Format.HTTP_HEADERS, response)
            self.span.finish()
            self.span = None
        return response


def hook(module):
    """ Hook method to install the Instana middleware into Django """
    if "INSTANA_DEV" in os.environ:
        print("==============================================================")
        print("Instana: Running django19 hook")
        print("==============================================================")

    if DJ19_INSTANA_MIDDLEWARE in module.settings.MIDDLEWARE_CLASSES:
        return

    if type(module.settings.MIDDLEWARE_CLASSES) is tuple:
        module.settings.MIDDLEWARE_CLASSES = (DJ19_INSTANA_MIDDLEWARE,) + module.settings.MIDDLEWARE_CLASSES
    elif type(module.settings.MIDDLEWARE_CLASSES) is list:
        module.settings.MIDDLEWARE_CLASSES = [DJ19_INSTANA_MIDDLEWARE] + module.settings.MIDDLEWARE_CLASSES
    else:
        print("Instana: Couldn't add InstanaMiddleware to Django")



import random
import os
import json
import time
import struct
import binascii
from instana import log

import sys
if sys.version_info.major is 2:
    string_types = basestring
else:
    string_types = str

_rnd = random.Random()
_current_pid = 0

BAD_ID_LONG = 3135097598  # Bad Cafe in base 10
BAD_ID_HEADER = "BADDCAFE"  # Bad Cafe


def generate_id():
    """ Generate a 64bit signed integer for use as a Span or Trace ID """
    global _current_pid

    pid = os.getpid()
    if (_current_pid != pid):
        _current_pid = pid
        _rnd.seed(int(1000000 * time.time()) ^ pid)
    return _rnd.randint(-9223372036854775808, 9223372036854775807)


def id_to_header(id):
    """ Convert a 64bit signed integer to an unsigned base 16 hex string """

    try:
        if not isinstance(id, int):
            return BAD_ID_HEADER

        byteString = struct.pack('>q', id)
        return str(binascii.hexlify(byteString).decode('UTF-8').lstrip('0'))
    except Exception as e:
        log.debug(e)
        return BAD_ID_HEADER


def header_to_id(header):
    """ Convert an unsigned base 16 hex string into a 64bit signed integer """

    if not isinstance(header, string_types):
        return BAD_ID_LONG

    try:
        # Test that header is truly a hexadecimal value before we try to convert
        int(header, 16)

        # Pad the header to 16 chars
        header = header.zfill(16)
        r = binascii.unhexlify(header)
        return struct.unpack('>q', r)[0]
    except ValueError:
        return BAD_ID_LONG


def to_json(obj):
    try:
        return json.dumps(obj, default=lambda obj: {k.lower(): v for k, v in obj.__dict__.items()},
                          sort_keys=False, separators=(',', ':')).encode()
    except Exception as e:
        log.info("to_json: ", e, obj)



import os
import sys
import socket
import subprocess
import threading as t
import fysom as f
from instana import log
import instana.agent_const as a


class Discovery(object):
    pid = 0
    name = None
    args = None
    fd = -1
    inode = ""

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    def to_dict(self):
        kvs = dict()
        kvs['pid'] = self.pid
        kvs['name'] = self.name
        kvs['args'] = self.args
        kvs['fd'] = self.fd
        kvs['inode'] = self.inode
        return kvs


class Fsm(object):
    RETRY_PERIOD = 30

    agent = None
    fsm = None
    timer = None

    def __init__(self, agent):
        log.info("Stan is on the scene.  Starting Instana instrumentation.")
        log.debug("initializing fsm")

        self.agent = agent
        self.fsm = f.Fysom({
            "events": [
                ("lookup",   "*",            "found"),
                ("announce", "found",        "announced"),
                ("ready",    "announced",    "good2go")],
            "callbacks": {
                "onlookup":       self.lookup_agent_host,
                "onannounce":     self.announce_sensor,
                "onready":        self.start_metric_reporting,
                "onchangestate":  self.printstatechange}})

        timer = t.Timer(2, self.fsm.lookup)
        timer.daemon = True
        timer.name = "Startup"
        timer.start()

    def printstatechange(self, e):
        log.debug('========= (%i#%s) FSM event: %s, src: %s, dst: %s ==========' %
                  (os.getpid(), t.current_thread().name, e.event, e.src, e.dst))

    def reset(self):
        self.fsm.lookup()

    def start_metric_reporting(self, e):
        self.agent.sensor.meter.run()

    def lookup_agent_host(self, e):
        host = a.AGENT_DEFAULT_HOST
        port = a.AGENT_DEFAULT_PORT

        if "INSTANA_AGENT_IP" in os.environ:
            host = os.environ["INSTANA_AGENT_IP"]
            if "INSTANA_AGENT_PORT" in os.environ:
                port = int(os.environ["INSTANA_AGENT_PORT"])

        elif self.agent.sensor.options.agent_host != "":
            host = self.agent.sensor.options.agent_host
            if self.agent.sensor.options.agent_port != 0:
                port = self.agent.sensor.options.agent_port

        h = self.check_host(host, port)
        if h == a.AGENT_HEADER:
            self.agent.set_host(host)
            self.agent.set_port(port)
            self.fsm.announce()
            return True
        elif os.path.exists("/proc/"):
            host = self.get_default_gateway()
            if host:
                h = self.check_host(host, port)
                if h == a.AGENT_HEADER:
                    self.agent.set_host(host)
                    self.agent.set_port(port)
                    self.fsm.announce()
                    return True

        log.info("Instana Host Agent couldn't be found. Scheduling retry.")
        self.schedule_retry(self.lookup_agent_host, e, "agent_lookup")
        return False

    def get_default_gateway(self):
        log.debug("checking default gateway")

        try:
            proc = subprocess.Popen(
                "/sbin/ip route | awk '/default/' | cut -d ' ' -f 3 | tr -d '\n'",
                shell=True, stdout=subprocess.PIPE)

            addr = proc.stdout.read()
            return addr.decode("UTF-8")
        except Exception as e:
            log.error(e)

            return None

    def check_host(self, host, port):
        log.debug("checking %s:%d" % (host, port))

        (_, h) = self.agent.request_header(
            self.agent.make_host_url(host, "/"), "GET", "Server")

        return h

    def announce_sensor(self, e):
        log.debug("announcing sensor to the agent")
        s = None
        pid = os.getpid()
        cmdline = []

        try:
            if os.path.isfile("/proc/self/cmdline"):
                with open("/proc/self/cmdline") as cmd:
                    cmdinfo = cmd.read()
                cmdline = cmdinfo.split('\x00')
            else:
                # Python doesn't provide a reliable method to determine what
                # the OS process command line may be.  Here we are forced to
                # rely on ps rather than adding a dependency on something like
                # psutil which requires dev packages, gcc etc...
                proc = subprocess.Popen(["ps", "-p", str(pid), "-o", "command"],
                                        stdout=subprocess.PIPE)
                (out, err) = proc.communicate()
                parts = out.split(b'\n')
                cmdline = [parts[1].decode("utf-8")]
        except Exception as err:
            cmdline = sys.argv
            log.debug(err)

        d = Discovery(pid=pid,
                      name=cmdline[0],
                      args=cmdline[1:])

        # If we're on a system with a procfs
        if os.path.exists("/proc/"):
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((self.agent.host, 42699))
            path = "/proc/%d/fd/%d" % (pid, s.fileno())
            d.fd = s.fileno()
            d.inode = os.readlink(path)

        (b, _) = self.agent.request_response(
            self.agent.make_url(a.AGENT_DISCOVERY_URL), "PUT", d)
        if b:
            self.agent.set_from(b)
            self.fsm.ready()
            log.info("Host agent available. We're in business. Announced pid: %i (true pid: %i)" %
                     (pid, self.agent.from_.pid))
            return True
        else:
            log.debug("Cannot announce sensor. Scheduling retry.")
            self.schedule_retry(self.announce_sensor, e, "announce")
        return False

    def schedule_retry(self, fun, e, name):
        log.debug("Scheduling: " + name)
        self.timer = t.Timer(self.RETRY_PERIOD, fun, [e])
        self.timer.daemon = True
        self.timer.name = name
        self.timer.start()
        log.debug('Threadlist: ', str(t.enumerate()))

    def test_agent(self, e):
        log.debug("testing communication with the agent")

        (b, _) = self.agent.head(self.agent.make_url(a.AGENT_DATA_URL))

        if not b:
            self.schedule_retry(self.test_agent, e, "agent test")
        else:
            self.fsm.test()



import threading as t
import socket
import time
import os
import instana

import opentracing.ext.tags as ext
from basictracer import Sampler, SpanRecorder
from .json_span import CustomData, Data, HttpData, JsonSpan, SDKData
from .agent_const import AGENT_TRACES_URL

import sys
if sys.version_info.major is 2:
    import Queue as queue
else:
    import queue


class InstanaRecorder(SpanRecorder):
    sensor = None
    registered_spans = ("django", "memcache", "rpc-client", "rpc-server", "urllib3", "wsgi")
    entry_kind = ["entry", "server", "consumer"]
    exit_kind = ["exit", "client", "producer"]
    queue = queue.Queue()

    def __init__(self, sensor):
        super(InstanaRecorder, self).__init__()
        self.sensor = sensor
        self.run()

    def run(self):
        """ Span a background thread to periodically report queued spans """
        self.timer = t.Thread(target=self.report_spans)
        self.timer.daemon = True
        self.timer.name = "Instana Span Reporting"
        self.timer.start()

    def report_spans(self):
        """ Periodically report the queued spans """
        while 1:
            if self.sensor.agent.can_send() and self.queue.qsize() > 0:
                url = self.sensor.agent.make_url(AGENT_TRACES_URL)
                self.sensor.agent.request(url, "POST", self.queued_spans())
            time.sleep(1)

    def queue_size(self):
        """ Return the size of the queue; how may spans are queued, """
        return self.queue.qsize()

    def queued_spans(self):
        """ Get all of the spans in the queue """
        spans = []
        while True:
            try:
                s = self.queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(s)
        return spans

    def clear_spans(self):
        """ Clear the queue of spans """
        self.queued_spans()

    def record_span(self, span):
        """
        Convert the passed BasicSpan into an JsonSpan and
        add it to the span queue
        """
        if self.sensor.agent.can_send() or "INSTANA_TEST" in os.environ:
            json_span = None

            if span.operation_name in self.registered_spans:
                json_span = self.build_registered_span(span)
            else:
                json_span = self.build_sdk_span(span)

            self.queue.put(json_span)

    def build_registered_span(self, span):
        """ Takes a BasicSpan and converts it into a registered JsonSpan """
        data = Data(http=HttpData(host=self.get_host_name(span),
                                  url=self.get_string_tag(span, ext.HTTP_URL),
                                  method=self.get_string_tag(span, ext.HTTP_METHOD),
                                  status=self.get_tag(span, ext.HTTP_STATUS_CODE)),
                    baggage=span.context.baggage,
                    custom=CustomData(tags=span.tags,
                    logs=self.collect_logs(span)))
        entityFrom = {'e': self.sensor.agent.from_.pid,
                      'h': self.sensor.agent.from_.agentUuid}

        return JsonSpan(
                    n=span.operation_name,
                    t=span.context.trace_id,
                    p=span.parent_id,
                    s=span.context.span_id,
                    ts=int(round(span.start_time * 1000)),
                    d=int(round(span.duration * 1000)),
                    f=entityFrom,
                    ec=self.get_tag(span, "ec"),
                    error=self.get_tag(span, "error"),
                    data=data)

    def build_sdk_span(self, span):
        """ Takes a BasicSpan and converts into an SDK type JsonSpan """

        custom_data = CustomData(tags=span.tags,
                                 logs=self.collect_logs(span))

        sdk_data = SDKData(name=span.operation_name,
                           custom=custom_data)

        sdk_data.Type = self.get_span_kind(span)
        data = Data(service=self.get_service_name(span), sdk=sdk_data)
        entityFrom = {'e': self.sensor.agent.from_.pid,
                      'h': self.sensor.agent.from_.agentUuid}

        return JsonSpan(
                    t=span.context.trace_id,
                    p=span.parent_id,
                    s=span.context.span_id,
                    ts=int(round(span.start_time * 1000)),
                    d=int(round(span.duration * 1000)),
                    n="sdk",
                    f=entityFrom,
                    data=data)

    def get_tag(self, span, tag):
        if tag in span.tags:
            return span.tags[tag]

        return None

    def get_string_tag(self, span, tag):
        ret = self.get_tag(span, tag)
        if not ret:
            return ""

        return ret

    def get_host_name(self, span):
        h = self.get_string_tag(span, "http.host")
        if len(h) > 0:
            return h

        h = socket.gethostname()
        if h and len(h) > 0:
            return h

        return "localhost"

    def get_service_name(self, span):
        return instana.service_name

    def get_span_kind(self, span):
        kind = ""
        if "span.kind" in span.tags:
            if span.tags["span.kind"] in self.entry_kind:
                kind = "entry"
            elif span.tags["span.kind"] in self.exit_kind:
                kind = "exit"
            else:
                kind = "local"
        return kind

    def collect_logs(self, span):
        logs = {}
        for l in span.logs:
            ts = int(round(l.timestamp * 1000))
            if ts not in logs:
                logs[ts] = {}

            for f in l.key_values:
                logs[ts][f] = l.key_values[f]

        return logs


class InstanaSampler(Sampler):

    def sampled(self, _):
        return False



from __future__ import absolute_import
import opentracing as ot
from basictracer.context import SpanContext
from instana import util, log


# The carrier can be a dict or a list.
# Using the trace header as an example, it can be in the following forms
# for extraction:
#   X-Instana-T
#   HTTP_X_INSTANA_T
#
# The second form above is found in places like Django middleware for
# incoming requests.
#
# For injection, we only support the standard format:
#   X-Instana-T


class HTTPPropagator():
    """A Propagator for Format.HTTP_HEADERS. """

    HEADER_KEY_T = 'X-Instana-T'
    HEADER_KEY_S = 'X-Instana-S'
    HEADER_KEY_L = 'X-Instana-L'
    ALT_HEADER_KEY_T = 'HTTP_X_INSTANA_T'
    ALT_HEADER_KEY_S = 'HTTP_X_INSTANA_S'
    ALT_HEADER_KEY_L = 'HTTP_X_INSTANA_L'

    def inject(self, span_context, carrier):
        try:
            trace_id = util.id_to_header(span_context.trace_id)
            span_id = util.id_to_header(span_context.span_id)

            if type(carrier) is dict or hasattr(carrier, "__dict__"):
                carrier[self.HEADER_KEY_T] = trace_id
                carrier[self.HEADER_KEY_S] = span_id
                carrier[self.HEADER_KEY_L] = "1"
            elif type(carrier) is list:
                carrier.append((self.HEADER_KEY_T, trace_id))
                carrier.append((self.HEADER_KEY_S, span_id))
                carrier.append((self.HEADER_KEY_L, "1"))
            else:
                raise Exception("Unsupported carrier type", type(carrier))

        except Exception as e:
            log.debug("inject error: ", str(e))

    def extract(self, carrier):  # noqa
        try:
            if type(carrier) is dict or hasattr(carrier, "__dict__"):
                dc = carrier
            elif type(carrier) is list:
                dc = dict(carrier)
            else:
                raise ot.SpanContextCorruptedException()

            # Look for standard X-Instana-T/S format
            if self.HEADER_KEY_T in dc and self.header_key_s in dc:
                trace_id = util.header_to_id(dc[self.HEADER_KEY_T])
                span_id = util.header_to_id(dc[self.HEADER_KEY_S])

            # Alternatively check for alternate HTTP_X_INSTANA_T/S style
            elif self.ALT_HEADER_KEY_T in dc and self.ALT_HEADER_KEY_S in dc:
                trace_id = util.header_to_id(dc[self.ALT_HEADER_KEY_T])
                span_id = util.header_to_id(dc[self.ALT_HEADER_KEY_S])

            return SpanContext(span_id=span_id,
                               trace_id=trace_id,
                               baggage={},
                               sampled=True)

        except Exception as e:
            log.debug("extract error: ", str(e))



from setuptools import setup, find_packages

setup(name='instana',
      version='0.7.8',
      download_url='https://github.com/instana/python-sensor',
      url='https://www.instana.com/',
      license='MIT',
      author='Instana Inc.',
      author_email='peter.lombardo@instana.com',
      description='Metrics sensor and trace collector for Instana',
      packages=find_packages(exclude=['tests', 'examples']),
      long_description="The instana package provides Python metrics and traces for Instana.",
      zip_safe=False,
      setup_requires=['nose>=1.0', 'flask>=0.12.2'],
      install_requires=['autowrapt>=1.0',
                        'fysom>=2.1.2',
                        'opentracing>=1.2.1,<1.3',
                        'basictracer>=2.2.0'],
      entry_points={'django': ['django.core.handlers.base = instana.django:hook'],
                    'django19': ['django.core.handlers.base = instana.django19:hook'],
                    'flask': ['flask = instana.flaskana:hook'],
                    'runtime': ['string = instana.runtime:hook']},
      test_suite='nose.collector',
      keywords=['performance', 'opentracing', 'metrics', 'monitoring'],
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking :: Monitoring',
        'Topic :: Software Development :: Libraries :: Python Modules'])



from __future__ import absolute_import
import instana
from instana.log import logger
import opentracing
import opentracing.ext.tags as ext
import wrapt


try:
    import urllib3 # noqa

    def collect(instance, args, kwargs):
        """ Build and return a fully qualified URL for this request """
        try:
            kvs = {}

            kvs['host'] = instance.host
            kvs['port'] = instance.port

            if args is not None and len(args) is 2:
                kvs['method'] = args[0]
                kvs['path'] = args[1]
            else:
                kvs['method'] = kwargs['method']
                kvs['path'] = kwargs['url']

            if type(instance) is urllib3.connectionpool.HTTPSConnectionPool:
                kvs['url'] = 'https://%s:%d%s' % (kvs['host'], kvs['port'], kvs['path'])
            else:
                kvs['url'] = 'http://%s:%d%s' % (kvs['host'], kvs['port'], kvs['path'])
        except Exception as e:
            logger.debug(e)
            return kvs
        else:
            return kvs

    @wrapt.patch_function_wrapper('urllib3', 'HTTPConnectionPool.urlopen')
    def urlopen_with_instana(wrapped, instance, args, kwargs):
        context = instana.internal_tracer.current_context()

        # If we're not tracing, just return
        if context is None:
            return wrapped(*args, **kwargs)

        try:
            span = instana.internal_tracer.start_span("urllib3", child_of=context)

            kvs = collect(instance, args, kwargs)
            if 'url' in kvs:
                span.set_tag(ext.HTTP_URL, kvs['url'])
            if 'method' in kvs:
                span.set_tag(ext.HTTP_METHOD, kvs['method'])

            instana.internal_tracer.inject(span.context, opentracing.Format.HTTP_HEADERS, kwargs["headers"])
            rv = wrapped(*args, **kwargs)

            span.set_tag(ext.HTTP_STATUS_CODE, rv.status)
            if 500 <= rv.status <= 599:
                span.set_tag("error", True)
                ec = span.tags.get('ec', 0)
                span.set_tag("ec", ec+1)

        except Exception as e:
            span.log_kv({'message': e})
            span.set_tag("error", True)
            ec = span.tags.get('ec', 0)
            span.set_tag("ec", ec+1)
            span.finish()
            raise
        else:
            span.finish()
            return rv

    instana.log.debug("Instrumenting urllib3")
except ImportError:
    pass



from __future__ import absolute_import
import os
import opentracing
from .sensor import Sensor
from .tracer import InstanaTracer
from .options import Options

if "INSTANA_DISABLE_AUTO_INSTR" not in os.environ:
    # Import & initialize instrumentation
    from .instrumentation import urllib3

"""
The Instana package has two core components: the sensor and the tracer.

The sensor is individual to each python process and handles process metric
collection and reporting.

The tracer upholds the OpenTracing API and is responsible for reporting
span data to Instana.
"""

__author__ = 'Instana Inc.'
__copyright__ = 'Copyright 2017 Instana Inc.'
__credits__ = ['Pavlo Baron', 'Peter Giacomo Lombardo']
__license__ = 'MIT'
__version__ = '0.7.8'
__maintainer__ = 'Peter Giacomo Lombardo'
__email__ = 'peter.lombardo@instana.com'

# For any given Python process, we only want one sensor as multiple would
# collect/report metrics in duplicate, triplicate etc..
#
# Usage example:
#
# import instana
# instana.global_sensor
#
global_sensor = Sensor(Options())

# The global OpenTracing compatible tracer used internally by
# this package.
#
# Usage example:
#
# import instana
# instana.internal_tracer.start_span(...)
#
internal_tracer = InstanaTracer()

# Set ourselves as the tracer.
opentracing.tracer = internal_tracer

# Optional application wide service name.
# Can be configured via environment variable or via code:
#
# export INSTANA_SERVICE_NAME=myservice
#   or
# instana.service_name = "myservice"
service_name = None

# User configurable EUM API key for instana.helpers.eum_snippet()
eum_api_key = ''

if "INSTANA_SERVICE_NAME" in os.environ:
    service_name = os.environ["INSTANA_SERVICE_NAME"]
